-- LearningAid v1.06 BETA 3 by Jamash (Kil'jaeden-US)

LearningAid = {}
LearningAid_Saved = {}

-- Copied from SpellBookFrame.lua
function LearningAid.UpdateButton(self)
  local id = self:GetID();
    
  local name = self:GetName();
  local iconTexture = getglobal(name.."IconTexture");
  local spellString = getglobal(name.."SpellName");
  local subSpellString = getglobal(name.."SubSpellName");
  local cooldown = getglobal(name.."Cooldown");
  local autoCastableTexture = getglobal(name.."AutoCastable");
  local highlightTexture = getglobal(name.."Highlight");
  local normalTexture = getglobal(name.."NormalTexture");
  if not LearningAid.inCombat then
    self:Enable();
  end

  if self.kind == BOOKTYPE_SPELL then 

    local texture = GetSpellTexture(id, BOOKTYPE_SPELL);

    -- If no spell, hide everything and return
    if ( not texture or (strlen(texture) == 0) ) then
      iconTexture:Hide();
      spellString:Hide();
      subSpellString:Hide();
      cooldown:Hide();
      autoCastableTexture:Hide();
      SpellBook_ReleaseAutoCastShine(self.shine)
      self.shine = nil;
      highlightTexture:SetTexture("Interface\\Buttons\\ButtonHilight-Square");
      self:SetChecked(0);
      normalTexture:SetVertexColor(1.0, 1.0, 1.0);
      return;
    end

    local start, duration, enable = GetSpellCooldown(id, BOOKTYPE_SPELL);
    CooldownFrame_SetTimer(cooldown, start, duration, enable);
    cooldown.start = start
    cooldown.duration = duration
    cooldown.enable = enable
    if ( enable == 1 ) then
      iconTexture:SetVertexColor(1.0, 1.0, 1.0);
    else
      iconTexture:SetVertexColor(0.4, 0.4, 0.4);
    end

    local spellName, subSpellName = GetSpellName(id, BOOKTYPE_SPELL);

    normalTexture:SetVertexColor(1.0, 1.0, 1.0);
    highlightTexture:SetTexture("Interface\\Buttons\\ButtonHilight-Square");
    spellString:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);

  -- Set Secure Action Button attribute
    if not LearningAid.inCombat then
      self:SetAttribute("spell*", spellName)
    end

    iconTexture:SetTexture(texture);
    spellString:SetText(spellName);
    subSpellString:SetText(subSpellName);
    if ( subSpellName ~= "" ) then
      spellString:SetPoint("LEFT", self, "RIGHT", 4, 4);
    else
      spellString:SetPoint("LEFT", self, "RIGHT", 4, 2);
    end
  elseif self.kind == "MOUNT" or self.kind == "CRITTER" then

    -- Some companions have two names, the display name and the spell name
    -- Make sure to use the spell name for casting
    local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(self.kind, id)
    local spellName = GetSpellInfo(creatureSpellID);
    iconTexture:SetTexture(icon)
    spellString:SetText(creatureName)
    subSpellString:SetText("")
    if not LearningAid.inCombat then
      self:SetAttribute("spell*", spellName)
    end
  end
  iconTexture:Show();
  spellString:Show();
  subSpellString:Show();
  --SpellButton_UpdateSelection(self);
end

-- Copied from SpellBookFrame.lua
function LearningAid.SpellButton_OnDrag(self) 
  local id = self:GetID();
  if self.kind == BOOKTYPE_SPELL then
    PickupSpell(id, self.kind);
  elseif self.kind == "MOUNT" or self.kind == "CRITTER" then
    PickupCompanion(self.kind, id)
  end
end

-- Copied from SpellBookFrame.lua
function LearningAid.SpellButton_OnEnter(self)
  local id = self:GetID();
  GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
  if self.kind == BOOKTYPE_SPELL then
    if GameTooltip:SetSpell(id, BOOKTYPE_SPELL) then
      self.UpdateTooltip = LearningAid.SpellButton_OnEnter
    else
      self.UpdateTooltip = nil
    end
  elseif self.kind == "MOUNT" or self.kind == "CRITTER" then
    local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(self.kind, id)
    if GameTooltip:SetHyperlink("spell:"..creatureSpellID) then
      self.UpdateTooltip = LearningAid.SpellButton_OnEnter;
    else
      self.UpdateTooltip = nil;
    end
  else
    DEFAULT_CHAT_FRAME:AddMessage("Invalid button type in LearningAid.SpellButton_OnEnter: "..self.kind)
  end
end

-- Copied from SpellBookFrame.lua
function LearningAid.SpellButton_UpdateSelection(self)
  if self.kind == BOOKTYPE_SPELL then
    local id = self:GetID()
    if IsSelectedSpell(id, BOOKTYPE_SPELL) then
      self:SetChecked("true");
    else
      self:SetChecked("false");
    end
  end
end

-- Copied from SpellBookFrame.lua
function LearningAid.SpellButton_OnModifiedClick(self, button) 
  local id = self:GetID()
  local spellName, subSpellName
  if self.kind == BOOKTYPE_SPELL then
    if ( id > MAX_SPELLS ) then
      return;
    end

    if ( IsModifiedClick("CHATLINK") ) then
      if ( MacroFrame and MacroFrame:IsShown() ) then
        spellName, subSpellName = GetSpellName(id, BOOKTYPE_SPELL);
          if ( spellName and not IsPassiveSpell(id, BOOKTYPE_SPELL) ) then
            if ( subSpellName and (strlen(subSpellName) > 0) ) then
              ChatEdit_InsertLink(spellName.."("..subSpellName..")");
            else
              ChatEdit_InsertLink(spellName);
            end
          end
        return;
      else
        local spellLink = GetSpellLink(id, BOOKTYPE_SPELL);
          if(spellLink) then
            ChatEdit_InsertLink(spellLink);
          end
        return;
      end
    end
    if ( IsModifiedClick("PICKUPACTION") ) then
      PickupSpell(id, BOOKTYPE_SPELL);
      return;
    end
  elseif self.kind == "MOUNT" or self.kind == "CRITTER" then
    local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(self.kind, id)
    if ( IsModifiedClick("CHATLINK") ) then
      if ( MacroFrame and MacroFrame:IsShown() ) then
        local spellName = GetSpellInfo(creatureSpellID);
        ChatEdit_InsertLink(spellName);
      else
        local spellLink = GetSpellLink(creatureSpellID)
        ChatEdit_InsertLink(spellLink);
      end
    elseif ( IsModifiedClick("PICKUPACTION") ) then
      LearningAid.SpellButton_OnDrag(self);
    end
  end
end

function LearningAid.OnLoad(self)
  self:RegisterEvent("SPELLS_CHANGED")
  self:RegisterEvent("PLAYER_LOGIN")
  self:RegisterEvent("COMPANION_LEARNED")
  --self:RegisterEvent("COMPANION_UPDATE")
  --self:RegisterEvent("TRADE_SKILL_SHOW")
  --self:RegisterEvent("TRADE_SKILL_CLOSE")
  --self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
  --self:RegisterEvent("CURRENT_SPELL_CAST_CHANGED")
  self:RegisterEvent("VARIABLES_LOADED")
  self:RegisterEvent("PLAYER_REGEN_DISABLED")
  self:RegisterEvent("PLAYER_REGEN_ENABLED")
  self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
  self:RegisterEvent("PLAYER_LEAVING_WORLD")
end

function LearningAid.OnEvent(self, event, ...)
  LearningAid.DebugPrint("Event "..event)
  local buttons = self.buttons
  if event == "PLAYER_REGEN_DISABLED" then
    --self:UnregisterEvent("SPELLS_CHANGED") -- do not attempt to update frame while in combat
    LearningAid.inCombat = true
    LearningAid.frame.closeButton:Disable()
  elseif event == "PLAYER_REGEN_ENABLED" then
    LearningAid.inCombat = false
    LearningAid.frame.closeButton:Enable()
    LearningAid.ProcessQueue()
    --local updated = LearningAid.DiffSpellBook()
    --if updated then
    --  LearningAid.UpdateSpellBook()
    --end
    --self:RegisterEvent("SPELLS_CHANGED")
  elseif (event == "SPELLS_CHANGED") and LearningAid.SpellBookCache ~= nil then
    if LearningAid.DiffSpellBook(LearningAid.retalenting) then
      LearningAid.UpdateSpellBook()
    else
      LearningAid.DebugPrint("Event "..event.." fired without spell changes")
    end
  elseif event == "COMPANION_LEARNED" then
    LearningAid.DiffCompanions()
    LearningAid.UpdateCompanions()
  elseif event == "COMPANION_UPDATE" then
    for i = 1, self.visible do
      local button = buttons[i]
      local kind = button.kind
      if kind == "MOUNT" or kind == "CRITTER" then
        local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, button:GetID())
        if isSummoned then
          button:SetChecked(true)
        else
          button:SetChecked(false)
        end
      end
    end
  elseif event == "TRADE_SKILL_SHOW" or event == "TRADE_SKILL_CLOSE" then
    for i = 1, self.visible do
      local button = buttons[i]
      if button.kind == BOOKTYPE_SPELL then
        if IsSelectedSpell(button:GetID(), button.kind) then
          button:SetChecked(true)
        else
          button:SetChecked(false)
        end
      end
    end
  elseif event == "SPELL_UPDATE_COOLDOWN" then
    for i = 1, self.visible do
      local button = buttons[i]
      if button.kind == BOOKTYPE_SPELL then
        LearningAid.UpdateButton(button)
      elseif button.kind == "MOUNT" or button.kind == "CRITTER" then
        local start, duration, enable = GetCompanionCooldown(button.kind, button:GetID())
        CooldownFrame_SetTimer(button.cooldown, start, duration, enable);
      end
    end
  elseif event == "CURRENT_SPELL_CAST_CHANGED" then
    for i = 1, self.visible do
      local button = buttons[i]
      if button.kind == BOOKTYPE_SPELL then
        LearningAid.SpellButton_UpdateSelection(button)
      end
    end
  elseif event == "UNIT_SPELLCAST_SUCCEEDED" then
    local unit, spell = ...
    if unit == "player" and (spell == LearningAid.activatePrimarySpec or spell == LearningAid.activateSecondarySpec) then
      LearningAid.retalenting = true
      --self:UnregisterEvent("SPELLS_CHANGED")
      self:RegisterEvent("PLAYER_TALENT_UPDATE")
    end
  elseif event == "PLAYER_TALENT_UPDATE" then
    self:UnregisterEvent("PLAYER_TALENT_UPDATE")
    LearningAid.retalenting = false
    --self:RegisterEvent("SPELLS_CHANGED")
    --if LearningAid.DiffSpellBook(true) then -- true flag means only remove spells, don't add new ones
    --  LearningAid.UpdateSpellBook()
    --end
  elseif event == "PLAYER_LEAVING_WORLD" then
    self:UnregisterEvent("SPELLS_CHANGED")
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
  elseif event == "PLAYER_ENTERING_WORLD" then
    --self:UnregisterEvent("PLAYER_ENTERING_WORLD")
    self:RegisterEvent("SPELLS_CHANGED")
  elseif event == "PLAYER_LOGIN" then
    LearningAid.UpdateSpellBook()
    LearningAid.UpdateCompanions()
  elseif event == "VARIABLES_LOADED" then
    if not LearningAid_Saved then LearningAid_Saved = {} end
    if LearningAid_Saved.locked then
      LearningAid.menuTable[1].text = LearningAid.unlockText
    else
      LearningAid_Saved.locked = false
    end
    if LearningAid_Saved.x and LearningAid_Saved.y then
      LearningAid.frame:ClearAllPoints()
      LearningAid.frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", LearningAid_Saved.x, LearningAid_Saved.y)
    end
  end
end

function LearningAid.UpdateCompanions()
  LearningAid.UpdateCompanionType("MOUNT")
  LearningAid.UpdateCompanionType("CRITTER")
end

function LearningAid.UpdateCompanionType(kind)
  LearningAid.companionCache[kind] = {}
  local cache = LearningAid.companionCache[kind]
  local i = 1
  local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, i)
  while creatureName do
    cache[i] = creatureName
    i = i + 1
    creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, i)
  end
end

function LearningAid.DiffCompanions()
  local updated
  updated = LearningAid.DiffCompanionType("MOUNT")
  if updated then
    LearningAid.UpdateCompanionType("MOUNT")
  else
    updated = LearningAid.DiffCompanionType("CRITTER")
    if updated then
      LearningAid.UpdateCompanionType("CRITTER")
    end
  end
end

function LearningAid.AddCompanion(kind, id)
  if LearningAid.inCombat then
    table.insert(LearningAid.queue, { action = "ADD", id = id, kind = kind})
  else
    LearningAid.LearnSpell(LearningAid.frame, kind, id)
    LearningAid.AddButton(LearningAid.frame, kind, id)
  end
end
function LearningAid.DiffCompanionType(kind)
  local i = 1
  local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, i)
  local cache = LearningAid.companionCache[kind]
  local updated = false
  while creatureName do 
    if cache[i] == nil or
       cache[i] ~= creatureName then
      LearningAid.DebugPrint("Found new companion, type "..kind..", index "..i)
      LearningAid.AddCompanion(kind, i)
      updated = true
      break
    end
    i = i + 1
    creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, i)
  end
  return updated
end

function LearningAid.UpdateSpellBook()
  LearningAid.SpellBookCache = {}
  local i = 1
  local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
  local spellAbsoluteID = LearningAid.AbsoluteSpellID(i)
  while spellName do
    LearningAid.SpellBookCache[i] = { name = spellName, rank = spellRank, absoluteID = spellAbsoluteID }
    i = i + 1
    spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
    spellAbsoluteID = LearningAid.AbsoluteSpellID(i)
  end
  LearningAid.DebugPrint("Updated Spellbook, "..i.." spells found")
end

function LearningAid.AddSpell(id)
  if LearningAid.inCombat then
    table.insert(LearningAid.queue, { action = "ADD", id = id, kind = BOOKTYPE_SPELL })
  else
    LearningAid.LearnSpell(LearningAid.frame, BOOKTYPE_SPELL, id)
    if not IsPassiveSpell(id, BOOKTYPE_SPELL) then -- if spell added that isn't passive
      -- Display button with draggable spell icon
      LearningAid.AddButton(LearningAid.frame, BOOKTYPE_SPELL, id)
    end
  end
end
function LearningAid.RemoveSpell(id)
  if LearningAid.inCombat then
    table.insert(LearningAid.queue, { action = "REMOVE", id = id, kind = BOOKTYPE_SPELL })
  else
    LearningAid.ClearButtonID(LearningAid.frame, BOOKTYPE_SPELL, id)
    LearningAid.ForgetSpell(LearningAid.frame, id)
  end
end
function LearningAid.ProcessQueue()
  if LearningAid.inCombat then
    LearningAid.DebugPrint("Cannot process action queue during combat in LearningAid.ProcessQueue")
    return
  end
  local queue = LearningAid.queue
  for index = 1, #queue do
    local item = queue[index]
    if item.kind == BOOKTYPE_SPELL then
      if item.action == "ADD" then
        LearningAid.AddSpell(item.id)
      elseif item.action == "REMOVE" then
        LearningAid.RemoveSpell(item.id)
      elseif item.action == "CLEAR" then
        LearningAid.ClearButtonID(LearningAid.frame, BOOKTYPE_SPELL, id)
      else
        LearningAid.DebugPrint("Invalid action type " .. item.action .. " in LearningAid.ProcessQueue")
      end
    elseif item.kind == "CRITTER" or item.kind == "MOUNT" then
      LearningAid.AddCompanion(item.kind, item.id)
    else
      LearningAid.DebugPrint("Invalid entry type " .. item.kind .. " in LearningAid.ProcessQueue")
    end
  end
  LearningAid.queue = {}
end
function LearningAid.DiffSpellBook(removeOnly)
  local i = 1
--  local offset = 0
  local cache = LearningAid.SpellBookCache
  local updated = false
  local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
  local spellAbsoluteID = LearningAid.AbsoluteSpellID(i)
  while spellName do
    if cache[i] == nil or
--       cache[i].name ~= spellName or
--       cache[i].rank ~= spellRank then
       cache[i].absoluteID ~= spellAbsoluteID then
      -- if spell removed
      if cache[i + 1] ~= nil and
--         cache[i + 1].name == spellName and
--         cache[i + 1].rank == spellRank then
         cache[i+1].absoluteID == spellAbsoluteID then
        LearningAid.DebugPrint("Old spell removed: "..cache[i].name.." ("..cache[i].rank..") id "..(i))
        LearningAid.RemoveSpell(i)
      elseif not removeOnly then
        LearningAid.DebugPrint("New spell found: "..spellName.." ("..spellRank..")") -- Old spell: "..cache[i + offset].name.." ("..cache[i + offset].rank..")")
        LearningAid.AddSpell(i)
      end
      -- remember to update cache
      updated = true
      break
    end
    i = i + 1
    spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL)
    spellAbsoluteID = LearningAid.AbsoluteSpellID(i)
  end
  -- if the last spell in the spellbook is removed
  if i == #cache and not updated then
    LearningAid.DebugPrint("Last spell removed: "..cache[i].name.." ("..cache[i].rank..") id "..i)
    LearningAid.RemoveSpell(i)
    updated = true
    --LearningAid.Button:Hide()
  end
  return updated
end
function LearningAid.LearnSpell(parent, kind, id)
  local buttons = parent.buttons
  for i = 1, parent.visible do
    local button = buttons[i]
    local buttonID = button:GetID()
    if button.kind == kind and buttonID >= id then
      button:SetID(buttonID + 1)
      LearningAid.UpdateButton(button)
    end
  end
end
function LearningAid.ForgetSpell(parent, id)
  local buttons = parent.buttons
  for i = 1, parent.visible do
    local button = buttons[i]
    local buttonID = button:GetID()
    if button.kind == BOOKTYPE_SPELL and buttonID > id then
      button:SetID(buttonID - 1)
      LearningAid.UpdateButton(button)
    end
  end
end

function LearningAid.CreateButton(parent)
  local count = #parent.buttons
  -- button global variable names start with "SpellButton" to work around an
  -- issue with the Blizzard Feedback Tool used in beta and on the PTR
  local name = "SpellButton_LearningAid_"..(count + 1)
  local button = CreateFrame("CheckButton", name, parent, "LearningAidSpellButtonTemplate")
  local subSpellName = getglobal(name.."SubSpellName")
  subSpellName:SetTextColor(NORMAL_FONT_COLOR.r - 0.1, NORMAL_FONT_COLOR.g - 0.1, NORMAL_FONT_COLOR.b - 0.1)
  parent.buttons[count + 1] = button
  button.index = count + 1
  if count > 0 then
    -- position relative to button above
    button:SetPoint("TOP", parent.buttons[count], "BOTTOM", 0, -LearningAid.buttonSpacing)
  else
    -- position relative to header
    button:SetPoint("TOPLEFT", parent.titleBar, "BOTTOMLEFT", 16, 0)
  end
  button:SetAttribute("type*", "spell")
  button:SetAttribute("type3", "hideButton")
  button:SetAttribute("alt-type*", "hideButton")
  button:SetAttribute("shift-type1", "linkSpell")
  button.hideButton = function(self, button, down)
    if not LearningAid.inCombat then
      LearningAid.ClearButtonIndex(self:GetParent(), self.index)
    end
  end
  button.linkSpell = LearningAid.SpellButton_OnModifiedClick
  button.iconTexture = _G[name.."IconTexture"]
  button.cooldown = _G[name.."Cooldown"]
  return button
end

function LearningAid.SpellButton_OnHide(self)
  LearningAid.DebugPrint("Hiding button "..self.index)
  self:SetChecked(false)
  self.iconTexture:SetVertexColor(1, 1, 1)
  self.cooldown:Hide()
end

function LearningAid.AddButton(frame, kind, id)
  local buttons = frame.buttons
  local button
  -- if bar is full
  if frame.visible == #buttons then
    button = LearningAid.CreateButton(frame)
    LearningAid.DebugPrint("Adding button id "..id.." index "..button.index)
  else
  -- if bar has free buttons
    button = buttons[frame.visible + 1]
    LearningAid.DebugPrint("Changing button index "..(frame.visible + 1).." from id "..button:GetID().." to "..id)
    button:Show()
  end

  button.kind = kind
  LearningAid.SetVisible(frame, frame.visible + 1)
  button:SetID(id)
  button:SetChecked(false)
  
  if kind == BOOKTYPE_SPELL then
    if id > 1 then
      local name, rank = GetSpellName(id, BOOKTYPE_SPELL)
      local prevName, prevRank = GetSpellName(id - 1, BOOKTYPE_SPELL)
      if name == prevName then
        LearningAid.DebugPrint("Found new rank of existing ability "..name.." "..prevRank)
        LearningAid.ClearButtonID(frame, kind, id - 1)
      else
        LearningAid.DebugPrint(name.." ~= "..prevName)
      end
    end
    if IsSelectedSpell(id, kind) then
      button:SetChecked(true)
    end
  elseif kind == "MOUNT" or kind == "CRITTER" then
    -- button.Companion = name
    local creatureID, creatureName, creatureSpellID, icon, isSummoned = GetCompanionInfo(kind, id)
    if isSummoned then
      button:SetChecked(true)
    end
  else
    DEFAULT_CHAT_FRAME:AddMessage("Invalid button type "..kind.." in LearningAid.AddButton()")
  end
  LearningAid.UpdateButton(button)
  frame:Show()
end

function LearningAid.ClearButtonID(frame, kind, id)
  local buttons = frame.buttons
  local i = 1
  -- not using a for loop because frame.visible may change during the loop execution
  while i <= frame.visible do
    if buttons[i].kind == kind and buttons[i]:GetID() == id then
      LearningAid.DebugPrint("Clearing button "..i.." with ID "..buttons[i]:GetID())
      LearningAid.ClearButtonIndex(frame, i)
    else
      LearningAid.DebugPrint("Button "..i.." has id "..buttons[i]:GetID().." which does not match "..id)
      i = i + 1
    end
  end
end

-- I have buttons 1 2 3 (4 5)
-- I remove button 2
-- I want 1 3 (3 4 5)
-- before, visible = 3
-- after, visible = 2
function LearningAid.ClearButtonIndex(frame, index)
  local buttons = frame.buttons
  for i = index, frame.visible - 1 do
    local button = buttons[i]
    local nextButton = buttons[i + 1]
    button:SetID(nextButton:GetID())
    button:SetChecked(nextButton:GetChecked())
    button.kind = nextButton.kind
    button.iconTexture:SetVertexColor(nextButton.iconTexture:GetVertexColor())
    local cooldown = button.cooldown
    local nextCooldown = nextButton.cooldown
    cooldown.start = nextCooldown.start
    cooldown.duration = nextCooldown.duration
    cooldown.enable = nextCooldown.enable
    if cooldown.start and cooldown.duration and cooldown.enable then 
      CooldownFrame_SetTimer(cooldown, cooldown.start, cooldown.duration, cooldown.enable)
    else
      cooldown:Hide()
    end
    --if buttons[i]:IsShown() then
    LearningAid.UpdateButton(button)
    --end
  end
  buttons[frame.visible]:Hide()
  LearningAid.SetVisible(frame, frame.visible - 1)
  if frame.visible == 0 then
    frame:Hide()
  end
end

function LearningAid.SetVisible(frame, visible)
  frame.visible = visible
  local top, left = frame:GetTop(), frame:GetLeft()
  frame:SetHeight(LearningAid.titleHeight + 10 + (LearningAid.buttonSize + LearningAid.buttonSpacing) * visible)
  frame:ClearAllPoints()
  frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", left, top)
end

function LearningAid.Hide(frame)
  if not LearningAid.inCombat then
    for i = 1, frame.visible do
      frame.buttons[i]:SetChecked(false)
      frame.buttons[i]:Hide()
    end
    frame.visible = 0
    frame:Hide()
  end
end

function LearningAid.TestAdd(kind, ...)
  DEFAULT_CHAT_FRAME:AddMessage("Testing!")
  local t = {...}
  for i = 1, #t do
    local id = t[i]
    if kind == BOOKTYPE_SPELL then
      if GetSpellInfo(id, kind) and not IsPassiveSpell(id, kind) then
        DEFAULT_CHAT_FRAME:AddMessage("Test: Adding spell id "..id)
        if LearningAid.inCombat then
          table.insert(LearningAid.queue, { action = "ADD", id = id, kind = kind })
        else
          LearningAid.AddButton(LearningAid.frame, kind, id)
        end
      else
        DEFAULT_CHAT_FRAME:AddMessage("Test: Spell id "..id.." is passive or does not exist")
      end
    elseif kind == "CRITTER" or kind == "MOUNT" then
      if GetCompanionInfo(kind, id) then
        DEFAULT_CHAT_FRAME:AddMessage("Test: Adding companion type "..kind.." id "..id)
        if LearningAid.inCombat then
          table.insert(LearningAid.queue, { action = "ADD", id = id, kind = kind})
        else
          LearningAid.AddButton(LearningAid.frame, kind, id)
        end
      else
        DEFAULT_CHAT_FRAME:AddMessage("Test: Companion type "..kind..", id "..id.." does not exist")
      end
    else
      DEFAULT_CHAT_FRAME:AddMessage("Test: Action type "..kind.." is not valid.  Valid types are SPELL, CRITTER or MOUNT.")
    end
  end
end

function LearningAid.TestRemove(...)
  DEFAULT_CHAT_FRAME:AddMessage("Testing!")
  local t = {...}
  for i = 1, #t do
    local id = t[i]
    DEFAULT_CHAT_FRAME:AddMessage("Test: Removing spell id "..id)
    if LearningAid.inCombat then
      table.insert(LearningAid.queue, { action = "CLEAR", id = id, kind = BOOKTYPE_SPELL })
    else
      LearningAid.ClearButtonID(LearningAid.frame, BOOKTYPE_SPELL, id)
    end
  end
end

function LearningAid.DebugPrint(msg)
  if LearningAid_Saved.debug then
    DEFAULT_CHAT_FRAME:AddMessage(msg)
  end
end

function LearningAid.OnShow(self)
  self:RegisterEvent("COMPANION_UPDATE")
  self:RegisterEvent("TRADE_SKILL_SHOW")
  self:RegisterEvent("TRADE_SKILL_CLOSE")
  self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
  self:RegisterEvent("CURRENT_SPELL_CAST_CHANGED")
end

function LearningAid.OnHide(self)
  self:UnregisterEvent("COMPANION_UPDATE")
  self:UnregisterEvent("TRADE_SKILL_SHOW")
  self:UnregisterEvent("TRADE_SKILL_CLOSE")
  self:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
  self:UnregisterEvent("CURRENT_SPELL_CAST_CHANGED")
end

function LearningAid.Lock(frame)
    LearningAid_Saved.locked = true
    LearningAid.menuTable[1].text = LearningAid.unlockText
end

function LearningAid.Unlock(frame)
    LearningAid_Saved.locked = false
    LearningAid.menuTable[1].text = LearningAid.lockText
end

function LearningAid.ToggleLock(frame)
  if LearningAid_Saved.locked then
    LearningAid.Unlock(frame)
  else
    LearningAid.Lock(frame)
  end
end


-- Transforms a spellbook ID into an absolute spell ID
function LearningAid.AbsoluteSpellID(id)
  local link = GetSpellLink(id, BOOKTYPE_SPELL)
  if link then
    local absoluteID = string.match(link, "Hspell:([^\124]+)\124")
    return absoluteID
  end
end

-- set addon variables
LearningAid.titleHeight = 40
LearningAid.width = 170
LearningAid.buttonSpacing = 5
LearningAid.buttonSize = 37
LearningAid.titleText = "Learning Aid"
LearningAid.lockText = "Lock Position"
LearningAid.unlockText = "Unlock Position"
LearningAid.closeText = "Close"
local version, build, date, tocversion = GetBuildInfo()
LearningAid.tocVersion = tocversion
LearningAid.companionCache = {}
LearningAid.menuHideDelay = 5
LearningAid.inCombat = false
LearningAid.retalenting = false
LearningAid.activatePrimarySpec = GetSpellInfo(63645)
LearningAid.activateSecondarySpec = GetSpellInfo(63644)
LearningAid.queue = {}

-- create main frame
local frame = CreateFrame("Frame", "LearningAid_Frame", UIParent)
LearningAid.frame = frame
frame:SetPoint("TOPLEFT", UIParent, "CENTER", 200, 100)
frame:SetWidth(LearningAid.width)
frame:SetHeight(LearningAid.titleHeight)
frame:SetMovable(true)
frame:SetClampedToScreen(true)
frame:SetScript("OnEvent", LearningAid.OnEvent)
frame:SetScript("OnShow", LearningAid.OnShow)
frame:SetScript("OnHide", LearningAid.OnHide)
frame.buttons = {}
frame.visible = 0
frame:Hide()
local backdrop = {
  bgFile = "Interface/DialogFrame/UI-DialogBox-Background",
  edgeFile = "Interface/DialogFrame/UI-DialogBox-Gold-Border",
  tile = false, tileSize = 16, edgeSize = 16,
  insets = { left = 4, right = 4, top = 4, bottom = 4 }
}
frame:SetBackdrop(backdrop)

-- create title bar
local titleBar = CreateFrame("Frame", "LearningAid_Frame_TitleBar", frame)
frame.titleBar = titleBar
titleBar:SetPoint("TOPLEFT")
titleBar:SetPoint("TOPRIGHT")
titleBar:SetHeight(LearningAid.titleHeight)
titleBar:RegisterForDrag("LeftButton")
titleBar:EnableMouse()
titleBar.text = titleBar:CreateFontString("LearningAid_Frame_Title_Text", "OVERLAY", "GameFontNormalLarge")
titleBar.text:SetText(LearningAid.titleText)
titleBar.text:SetPoint("CENTER", titleBar, "CENTER", 0, 0)

-- create close button
local closeButton = CreateFrame("Button", "LearningAid_Frame_CloseButton", titleBar)
frame.closeButton = closeButton
closeButton:SetWidth(32)
closeButton:SetHeight(32)
closeButton:SetPoint("RIGHT", titleBar, "RIGHT", -2, 0)
closeButton:SetNormalTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Up")
closeButton:SetPushedTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Down")
closeButton:SetDisabledTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Disabled")
closeButton:SetHighlightTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Highlight")
closeButton:SetScript("OnClick", function () LearningAid.Hide(frame) end)

-- initialize right-click menu
LearningAid.menuTable = {
  { text = LearningAid.lockText, 
    func = function () LearningAid.ToggleLock(frame) end },
  { text = LearningAid.closeText,
    func = function () LearningAid.Hide(frame) end }
}

local Menu = CreateFrame("Frame", "LearningAid_Menu", titleBar, "UIDropDownMenuTemplate")

-- set drag and click handlers for the title bar
titleBar:SetScript(
  "OnDragStart",
  function (self, button)
    if not LearningAid_Saved.locked then
      self:GetParent():StartMoving()
    end
  end
)

titleBar:SetScript(
  "OnDragStop",
  function (self)
    local parent = self:GetParent()
    parent:StopMovingOrSizing()
    LearningAid_Saved.x = parent:GetLeft()
    LearningAid_Saved.y = parent:GetTop()
  end
)

titleBar:SetScript(
  "OnMouseUp",
  function (self, button)
    if button == "MiddleButton" then
      LearningAid.Hide(self:GetParent())
    elseif button == "RightButton" then
      EasyMenu(LearningAid.menuTable, Menu, titleBar, 0, 8, "MENU", LearningAid.menuHideDelay)
    end
  end
)

-- set up the slash command
function LearningAid.SlashCommand(msg)
  local pattern = "(%S+)%s*"
  if msg then
    local first, last, text = strfind(msg, pattern)
    if text then
      local lowertext = strlower(text)
    
      if lowertext == "debug" then
        first, last, text = strfind(msg, pattern, last + 1)
        if text then
          lowertext = strlower(text)
          if lowertext == "0" or lowertext == "off" then
            LearningAid_Saved.debug = false
          else
            LearningAid_Saved.debug = true
          end
        else
          DEFAULT_CHAT_FRAME:AddMessage(LearningAid_Saved.debug and "on" or "off")
        end
      elseif lowertext == "hide" then
        LearningAid.Hide(frame)
      elseif lowertext == "lock" then
        LearningAid.Lock(frame)
      elseif lowertext == "unlock" then
        LearningAid.Unlock(frame)
      elseif lowertext == "test" then
        first, last, text = strfind(msg, pattern, last + 1)
        if text then
          lowertext = strlower(text)
          local arguments = {}
          for item in string.gmatch(string.sub(msg, last + 1), "%S+") do
            if string.match(item, "^%d+$") then
              item = tonumber(item)
            end
            tinsert(arguments, item)
          end
          if lowertext == "add" then
            LearningAid.TestAdd(unpack(arguments))
          elseif lowertext == "remove" then
            LearningAid.TestRemove(unpack(arguments))
          else
            DEFAULT_CHAT_FRAME:AddMessage("Invalid argument(s) to command 'test' in "..msg)
          end
        end
      end
    end
  end
end

SlashCmdList["LearningAid"] = LearningAid.SlashCommand
SLASH_LearningAid1 = "/learningaid"
SLASH_LearningAid2 = "/la"

-- initialize the addon
LearningAid.OnLoad(frame)
